# pylint: disable=no-name-in-module

from __future__ import annotations

import os
from abc import ABCMeta, abstractmethod
from os import PathLike
from pathlib import Path
from types import ModuleType
from typing import (  # noqa: I101
    TYPE_CHECKING,
    Any,
    ClassVar,
    Dict,
    Optional,
    Sequence,
    Tuple,
    Type,
    Union,
    no_type_check,
)
from warnings import warn

import pydantic
from pydantic import BaseModel
from pydantic.env_settings import InitSettingsSource
from pydantic.fields import ModelField
from pydantic.typing import get_origin, is_union
from pydantic.utils import lenient_issubclass

from ._optional_libs import dotenv  # noqa: I202
from ._optional_libs import hcl2, toml, yaml
from .dumploads import (
    ConfigLoadCallable,
    hcl2_load,
    json_load,
    toml_load,
    yaml_load,
)


if TYPE_CHECKING:
    from ipl_config import BaseSettings  # pragma: no cover


class SettingsStrategyMetaclass(ABCMeta):  # noqa: B024
    @no_type_check
    def __call__(cls, *args, **kwargs):
        """
        Ensure the class has all its dependencies during constructor
        without needs to call to `super`
        """
        for dep in getattr(cls, '__dependencies__', None) or ():
            if isinstance(dep, Exception):
                raise dep
        return super().__call__(*args, **kwargs)


# type: ignore[too-few-public-methods]
# pylint: disable=too-few-public-methods
class SettingsStrategy(metaclass=SettingsStrategyMetaclass):
    __dependencies__: ClassVar[
        Optional[Sequence[Union[ModuleType, ImportError]]]
    ] = None

    @abstractmethod
    def __call__(
        self, clazz: Union[Type[BaseSettings], BaseSettings]
    ) -> Dict[str, Any]:
        pass  # pragma: no cover


# pylint: disable=too-few-public-methods
class EnvSettingsStrategy(SettingsStrategy):
    __slots__ = 'env_prefix', 'env_vars', 'case_sensitive'

    def __init__(
        self,
        env_prefix: Optional[str] = None,
        env_vars: Union[Dict[str, Optional[str]], None] = None,
        case_sensitive: Optional[bool] = False,
    ):
        self.env_prefix: Optional[str] = env_prefix
        self.env_vars: Dict[str, Optional[str]] = (
            env_vars  # type: ignore[assignment]
            if env_vars is not None
            else os.environ
        )
        self.case_sensitive: Optional[bool] = case_sensitive

        # self.env_vars: Dict[str, Optional[str]] = {
        #     **os.environ,
        #     **(read_env_file(
        #         env_file, encoding=env_file_encoding
        #     ) if env_file else {}),
        # }

        if not case_sensitive:
            if env_prefix:
                self.env_prefix = env_prefix.lower()
            self.env_vars = {k.lower(): v for k, v in self.env_vars.items()}

    def __call__(
        self,
        clz: Union[Type[BaseModel], BaseModel],
        prefix: Optional[str] = None,
    ) -> Dict[str, Any]:
        if prefix is None:
            prefix = self.env_prefix

        return dict(
            filter(
                lambda x: x is not None,  # type: ignore
                (
                    self._get_env_val(clz=clz, field=f, prefix=prefix)
                    for f in clz.__fields__.values()
                ),
            )
        )

    def _get_env_val(  # pylint: disable=too-many-branches
        self,
        clz: Union[Type[BaseModel], BaseModel],
        field: ModelField,
        prefix: Optional[str] = None,
    ) -> Optional[Tuple[str, Any]]:
        prefix = prefix or ''

        if field.field_info.extra.get('deprecated'):
            warn(f"{field.name!r} is deprecated", DeprecationWarning)
        if field.has_alias:
            warn('Instead of aliases use the `env` setting', FutureWarning)

        env_name = field.field_info.extra.get('env')
        if not env_name:
            env_name = prefix + (prefix and '_' or '') + field.name
        if not self.case_sensitive:
            env_name = env_name.lower()

        env_val: Any = self.env_vars.get(env_name)

        if (
            field.shape == pydantic.fields.SHAPE_SINGLETON
            and lenient_issubclass(field.type_, BaseModel)  # noqa: W503
        ):
            env_val = self(field.type_, prefix=env_name)
        elif field.is_complex():
            if env_val is not None:
                env_val = clz.__config__.json_loads(env_val)
        elif is_union(get_origin(field.type_)):
            for f in field.sub_fields:
                try:
                    if env_val is not None and f.is_complex():
                        env_val = clz.__config__.json_loads(env_val)
                    else:
                        break
                except ValueError:
                    pass

        if env_val is not None:
            return field.alias, env_val

        return None


# pylint: disable=too-few-public-methods
class DotEnvSettingsStrategy(EnvSettingsStrategy):
    __slots__ = ()

    def __init__(
        self,
        env_prefix: Optional[str] = None,
        case_sensitive: Optional[bool] = False,
        env_file: Union[str, PathLike, None] = None,
        env_file_encoding: Optional[str] = None,
    ):
        super().__init__(
            env_prefix=env_prefix,
            case_sensitive=case_sensitive,
            env_vars=read_env_file(env_file, encoding=env_file_encoding)
            if env_file
            else {},
        )


# pylint: disable=too-few-public-methods
class KwSettingsStrategy(SettingsStrategy, InitSettingsSource):
    def __call__(
        self, clazz: Union[Type[BaseSettings], BaseSettings]
    ) -> Dict[str, Any]:
        return self.init_kwargs

    def __init__(self, **kw: Any) -> None:
        super().__init__(init_kwargs=kw)


class FileSettingsStrategy(SettingsStrategy):
    __slots__ = 'path', 'config_format'

    __extensions__: ClassVar[Sequence[str]] = ()

    def __init__(
        self, path: Union[str, PathLike], config_format: Optional[str] = None
    ):
        self.path: Path = Path(path).expanduser()
        self.config_format: Optional[str] = config_format

    def __call__(
        self, clazz: Union[Type[BaseSettings], BaseSettings]
    ) -> Dict[str, Any]:
        if not self.is_acceptable(self.path, self.config_format):
            return {}

        loader = self.get_loader(clazz)
        return loader(self.path)

    @abstractmethod
    def get_loader(
        self, clazz: Union[Type[BaseSettings], BaseSettings]
    ) -> ConfigLoadCallable:
        pass  # pragma: no cover

    @classmethod
    def is_acceptable(
        cls, path: Optional[Path] = None, config_format: Optional[str] = None
    ) -> bool:
        """
        :return: True if format is acceptable to loads
        """
        if path is None:
            return False

        if config_format is None:
            config_format = cls.detect_format(path)

        return bool(config_format and config_format in cls.__extensions__)

    @classmethod
    def detect_format(cls, path: Path) -> Optional[str]:
        return path.suffix[1:]


class JsonSettingsStrategy(FileSettingsStrategy):
    __extensions__ = 'json', 'js'

    def get_loader(
        self, clazz: Union[Type[BaseSettings], BaseSettings]
    ) -> ConfigLoadCallable:
        return json_load


class YamlSettingsStrategy(FileSettingsStrategy):
    __dependencies__ = (yaml,)
    __extensions__ = 'yaml', 'yml'

    def get_loader(
        self, clazz: Union[Type[BaseSettings], BaseSettings]
    ) -> ConfigLoadCallable:
        return yaml_load


class TomlSettingsStrategy(FileSettingsStrategy):
    __dependencies__ = (toml,)
    __extensions__ = 'toml', 'tml'

    def get_loader(
        self, clazz: Union[Type[BaseSettings], BaseSettings]
    ) -> ConfigLoadCallable:
        return toml_load


class Hcl2SettingsStrategy(FileSettingsStrategy):
    __dependencies__ = (hcl2,)
    __extensions__ = 'hcl', 'hcl2', 'tf'

    def get_loader(
        self, clazz: Union[Type[BaseSettings], BaseSettings]
    ) -> ConfigLoadCallable:
        return hcl2_load


def read_env_file(
    path: Union[str, PathLike], *, encoding: Optional[str] = None
) -> Dict[str, Optional[str]]:
    path = Path(path).expanduser()
    is_env_default = str(path) == '.env'
    is_env_exists = path.is_file()

    if is_env_exists and isinstance(dotenv, Exception):
        warn(str(dotenv), ImportWarning)
    elif not is_env_exists and not is_env_default:
        warn(f"{str(path)!r} is not a file", UserWarning)
    else:
        return dotenv.dotenv_values(path, encoding=encoding or 'utf-8')

    return {}



# pylint: disable=no-name-in-module

import sys
from collections import OrderedDict
from decimal import Decimal
from os import PathLike
from pathlib import Path
from typing import (
    AbstractSet,
    Any,
    ClassVar,
    Dict,
    Generator,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Type,
    Union,
)

from pydantic import BaseConfig, BaseModel
from pydantic.config import Extra
from pydantic.utils import deep_update

from .dumploads import StrPathIO, json_dump, toml_dump, yaml_dump
from .source import (
    DotEnvSettingsStrategy,
    EnvSettingsStrategy,
    Hcl2SettingsStrategy,
    JsonSettingsStrategy,
    KwSettingsStrategy,
    SettingsStrategy,
    TomlSettingsStrategy,
    YamlSettingsStrategy,
)


IntStr = Union[int, str]
AbstractSetIntStr = AbstractSet[IntStr]
MappingIntStrAny = Mapping[IntStr, Any]
TupleGenerator = Generator[Tuple[str, Any], None, None]


class BaseSettings(BaseModel):
    __slots__ = ()

    class Config(BaseConfig):  # pylint: disable=too-few-public-methods
        env_prefix: Optional[str] = 'APP'
        env_file: Union[str, PathLike, None] = '.env'
        env_file_encoding: Optional[str] = None
        case_sensitive: bool = False
        validate_all: bool = True
        extra: Extra = Extra.ignore
        arbitrary_types_allowed = True

    __config__: ClassVar[Type[Config]] = Config

    def __init__(  # pylint: disable=too-many-arguments
        self,
        env_prefix: Optional[str] = None,
        env_file: Union[str, PathLike, None] = None,
        config_file: Union[str, PathLike, None] = None,
        config_format: Optional[str] = None,
        source_strategies: Optional[Sequence[SettingsStrategy]] = None,
        **kw: Any,
    ) -> None:
        if config_file is not None:
            config_file = Path(config_file)

        if source_strategies is None:
            cfg = self.__config__
            source_strategies = [
                KwSettingsStrategy(**kw),
                EnvSettingsStrategy(
                    env_prefix=env_prefix or cfg.env_prefix,
                    case_sensitive=cfg.case_sensitive,
                ),
                DotEnvSettingsStrategy(
                    env_prefix=env_prefix or cfg.env_prefix,
                    env_file=env_file or cfg.env_file,
                    env_file_encoding=cfg.env_file_encoding,
                    case_sensitive=cfg.case_sensitive,
                ),
            ]
            if config_file:
                for s in (
                    JsonSettingsStrategy,
                    YamlSettingsStrategy,
                    TomlSettingsStrategy,
                    Hcl2SettingsStrategy,
                ):
                    if s.is_acceptable(config_file, config_format):
                        source_strategies.append(
                            s(
                                config_file,
                                config_format,  # type: ignore[abstract]
                            )
                        )
                        break
                else:
                    raise NotImplementedError(
                        f"No readers found for the config file: {config_file}"
                    )

        super().__init__(
            **deep_update(*reversed([s(self) for s in source_strategies]))
        )

    def safe_dict(
        self,
        *,
        include: Optional[Union[AbstractSetIntStr, MappingIntStrAny]] = None,
        exclude: Optional[Union[AbstractSetIntStr, MappingIntStrAny]] = None,
        by_alias: bool = False,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
    ) -> Dict[str, Any]:
        return dict(
            self.safe_iter(
                to_dict=True,
                by_alias=by_alias,
                include=include,
                exclude=exclude,
                exclude_unset=exclude_unset,
                exclude_defaults=exclude_defaults,
                exclude_none=exclude_none,
            )
        )

    def safe_iter(  # pylint: disable=too-many-arguments
        self,
        to_dict: bool = False,
        by_alias: bool = False,
        include: Optional[Union[AbstractSetIntStr, MappingIntStrAny]] = None,
        exclude: Optional[Union[AbstractSetIntStr, MappingIntStrAny]] = None,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
    ) -> TupleGenerator:
        for k, v in super()._iter(
            to_dict,
            by_alias,
            include,
            exclude,
            exclude_unset,
            exclude_defaults,
            exclude_none,
        ):
            try:
                v = self.__json_encoder__(v)
            except TypeError:
                pass
            yield k, v,

    # def load_ini(self):
    #     from configparser import ConfigParser
    #     ConfigParser().read_file()

    def _write_json(self, o: Dict[str, Any], f: StrPathIO, **kw: Any) -> None:
        # take encoder from `self.__config__.json_encoders`
        encoder = kw.pop('default', self.__json_encoder__)
        json_dump(o, f, default=encoder, **kw)

    def write_schema(self, f: StrPathIO = sys.stdout, **kw: Any) -> None:
        return self._write_json(self.schema(), f, **kw)

    def write_json(self, f: StrPathIO = sys.stdout, **kw: Any) -> None:
        return self._write_json(self.dict(), f, **kw)

    def write_toml(self, f: StrPathIO = sys.stdout, **kw: Any) -> None:
        return toml_dump(self.safe_dict(), f, **kw)

    def write_yaml(self, f: StrPathIO = sys.stdout, **kw: Any) -> None:
        return yaml_dump(self.safe_dict(), f, **kw)

    def to_env(self, **kw: Any) -> Dict[str, str]:
        prefix = self.__config__.env_prefix
        if not prefix:
            return self._to_env(self.dict(**kw))
        return self._to_env(self.dict(**kw), prefix)

    def _to_env(self, ns: Dict[str, Any], *path: str) -> Dict[str, str]:
        res = OrderedDict()

        for name, model in ns.items():
            _path = path + (name,)
            env_name = '_'.join(_.upper() for _ in _path)

            if not isinstance(model, dict):
                # for mypy
                env_val: Union[bool, bytes, str, int, float, complex, Decimal]
                if model is None:
                    env_val = ''
                elif isinstance(model, bool):
                    env_val = int(model)
                elif isinstance(model, bytes):
                    env_val = model.decode(
                        encoding='utf-8', errors='surrogateescape'
                    )
                elif isinstance(model, (str, int, float, complex, Decimal)):
                    env_val = model
                else:
                    raise NotImplementedError(
                        f"Not a scalar: {env_name}={type(model)}"
                    )

                res[env_name] = str(env_val)
            else:
                for k, v in self._to_env(model, *_path).items():
                    res[k] = v

        return res



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ipl_config']

package_data = \
{'': ['*']}

install_requires = \
['pydantic>=1.10.2,<2.0.0']

extras_require = \
{':python_version < "3.8"': ['typing-extensions>=4.3.0,<5.0.0'],
 'dotenv': ['python-dotenv>=0.21.0,<0.22.0'],
 'hcl2': ['python-hcl2>=3.0.5,<4.0.0'],
 'toml': ['toml>=0.10.2,<0.11.0'],
 'yaml': ['pyyaml>=5.0,<7.0']}

setup_kwargs = {
    'name': 'ipl-config',
    'version': '0.1.10',
    'description': 'InPlat config adapters',
    'long_description': '[![tests](https://github.com/koi8-r/ipl-config/actions/workflows/ci.yml/badge.svg)](https://github.com/koi8-r/ipl-config/actions/workflows/ci.yml)\n[![codecov](https://codecov.io/gh/koi8-r/ipl-config/branch/master/graph/badge.svg?token=OKURU75Y7A)](https://codecov.io/gh/koi8-r/ipl-config)\n[![pypi](https://img.shields.io/pypi/v/ipl-config.svg)](https://pypi.python.org/pypi/ipl-config)\n[![versions](https://img.shields.io/pypi/pyversions/ipl-config.svg)](https://github.com/koi8-r/ipl-config)\n\n\n# Config adapters with pydantic behavior\n- json\n- yaml\n- toml\n- hcl2\n- environ\n- .env\n- TODO: multiline PEM keys load with cryptography\n\n## Examples\n### .env\n```dotenv\nAPP_VERSION=v0.0.1a1\nAPP_HTTP_HOST=myname.lan\nHTTP_2=true\nAPP_PRIVATE_KEY="-----BEGIN PRIVATE KEY-----\nMIGHAgEAMBMGByqGSM49AgEGCCqGSM49AwEHBG0wawIBAQQg8M4vd2AmKTW0/nqc\nYQBi/bRZjkVezdGHi+zH5kYvm/2hRANCAATxEs1e8hqwpYCTk3amfq/UnGyvViPZ\nMidz4nFzQvcq7A9Ju/wvEfLDjA131kh2Sk+x3dgLxhTf7yKJXZC0jg3d\n-----END PRIVATE KEY-----"\n```\n### config.yaml\n```yaml\nhttp:\n  port: 10001\n  transport:\n    timeout: 60.0\n    buffer_size: 65535\n  interfaces:\n    - 127.0.0.1\n    - 192.168.0.1\nversion: 1\n```\n### yaml with env, dotenv and args overrides\n```python\nfrom datetime import datetime\nfrom ipaddress import IPv4Address\nfrom os import environ\nfrom pathlib import Path\nfrom typing import Dict, Union\n\nfrom pydantic import BaseModel, Field\n\nfrom ipl_config import BaseSettings\n\n\nclass TcpTransport(BaseModel):\n    timeout: float  # from config file\n    buffer_size: int = Field(0.01, env=\'BUFF_SIZE\')  # from env\n\n\nclass Http(BaseModel):\n    host: str  # from dotenv\n    bind: str  # from env\n    port: int  # from config file\n    interfaces: list[IPv4Address]  # from config file\n    transport: TcpTransport\n    http2: bool = Field(env=\'HTTP_2\')  # from dotenv\n\n\nclass IplConfig(BaseSettings):\n    version: str  # from kwargs\n    created: datetime  # from env\n    http: Http  # env also works for complex objects\n    private_key: str  # from dotenv\n    group_by_id: Union[Dict[int, str], None]\n\n\nif __name__ == "__main__":\n    environ[\'app_http_bind\'] = \'1.1.1.1\'\n    environ[\'buff_size\'] = \'-1\'\n    environ[\'app_created\'] = \'2000-01-01T00:00:00Z\'\n    environ[\'app_group_by_id_0\'] = \'root\'\n\n\n    root = Path(\'.\')\n\n    cfg = IplConfig(\n        version=\'v0.0.1a1\',\n        env_file=root / \'.env\',\n        config_file=root / \'config.yaml\',\n    )\n    cfg.write_json(indent=4)\n    print()\n```\n',
    'author': 'InPlat',
    'author_email': 'None',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/koi8-r/ipl-config',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7.2,<4.0.0',
}


setup(**setup_kwargs)
